# -*- coding: utf-8 -*-
"""Windows configuration utilities."""

import os
from contextlib import contextmanager
from io import BytesIO, StringIO
from logging import DEBUG, getLogger
from pathlib import Path
from re import sub

# pywin32 only needs to be installed on windows
try:
    from win32crypt import CryptProtectData, CryptUnprotectData
    from win32security import (
        DACL_SECURITY_INFORMATION,
        GetFileSecurity,
        LookupAccountSid,
        SetFileSecurity,
    )
except ImportError:
    CryptProtectData = None
    CryptUnprotectData = None
    DACL_SECURITY_INFORMATION = None
    GetFileSecurity = None
    LookupAccountSid = None
    SetFileSecurity = None


FILE_WRITE_DATA = 1 << 1
GENERIC_ALL = 1 << 28


@contextmanager
def open_dpapi(path, method="r"):
    """Opens and decrypts or encrypts the specified file.

    Arguments:
        path (str): Path to file.
        method (str): "r" or "w" to read or write file (default "r").

    Yields:
        Stream object to read or write.
    """
    method = "wb" if method == "w" else "rb"
    description = sub(r"\..*", "", os.path.basename(path))
    if "WireSock" in str(path):
        description = ""
    file = open(path, method)
    buffer = StringIO()

    try:
        if method == "rb":
            decrypt(file, buffer, description)
            buffer.seek(0)
        yield buffer

    finally:
        if method == "wb":
            buffer.seek(0)
            encrypt(buffer, file, description)
        file.close()


def decrypt(cipher_stream, plain_stream=None, description=""):
    """Decrypts the first stream to the second.

    Arguments:
        cipher_stream: Input stream with encrypted bytes.
        plain_stream: Output stream for decrypted text.
        description (str): Expected description included in encrypted bytes.

    Returns:
        Output stream.

    Raises:
        ValueError: Expected description does not match actual description.
    """
    if not plain_stream:
        plain_stream = StringIO()

    cipher_bytes = cipher_stream.read()
    actual_description, plain_bytes = CryptUnprotectData(cipher_bytes)
    if description and description != actual_description:
        msg = f"expected description '{description}'; actual '{actual_description}'"
        raise ValueError(msg)
    plain_stream.write(plain_bytes.decode("utf-8"))

    return plain_stream


def encrypt(plain_stream, cipher_stream=None, description=""):
    """Encrypts the first stream to the second.

    Arguments:
        plain_stream: Input stream with decrypted text.
        cipher_stream: Output stream for encrypted bytes.
        description (str): Description to include in encrypted bytes.

    Returns:
        Output stream.
    """
    if not cipher_stream:
        cipher_stream = BytesIO()

    plain_text = plain_stream.read()
    cipher_bytes = CryptProtectData(plain_text.encode("utf-8"), description)
    cipher_stream.write(cipher_bytes)

    return cipher_stream


def restrict_access_to_cnf_dir(cnf):
    """Restricts permissions of the specified config file directory.

    Arguments:
        cnf: Config object.
    """
    cnf_file = Path(cnf.cnf_file)
    if not cnf_file.is_file():
        getLogger(__name__).error("cnf file not found: %s", cnf_file)
        return

    cnf_dir = cnf_file.parent
    restrict_access_to_file(cnf_dir)
    for file in cnf_dir.iterdir():
        restrict_access_to_file(file)


# simpler for now to keep this logic together rather than subdivide it more functions
def restrict_access_to_file(file):  # noqa: CCR001
    """Restricts permissions of the specified file.

    Arguments:
        file (str): Path to file.
    """
    file = str(file)
    getLogger(__name__).debug("restricting access to %s", file)

    descriptor = GetFileSecurity(file, DACL_SECURITY_INFORMATION)
    dacl = descriptor.GetSecurityDescriptorDacl()

    keep = set()
    for i in range(0, dacl.GetAceCount()):
        (ace_type, flags), mask, sid = dacl.GetAce(i)
        if ace_type == 0 and ((mask & FILE_WRITE_DATA) or (mask & GENERIC_ALL)):
            _debug_keep_ace(sid, mask)
            keep.add(str(sid))

    for i in reversed(range(0, dacl.GetAceCount())):
        (ace_type, flags), mask, sid = dacl.GetAce(i)
        if ace_type == 0 and str(sid) not in keep:
            _debug_remove_ace(sid, mask)
            dacl.DeleteAce(i)

    descriptor.SetSecurityDescriptorDacl(1, dacl, 0)
    SetFileSecurity(file, DACL_SECURITY_INFORMATION, descriptor)


def _debug_keep_ace(sid, mask):
    """Logs info about ACE being kept.

    Arguments:
        sid: SID object.
        mask: Rights.
    """
    logger = getLogger(__name__)
    if logger.isEnabledFor(DEBUG):
        formatted_sid = _format_ace_sid(sid)
        formatted_mask = _format_ace_mask(mask)
        logger.debug("keep access for %s (%s)", formatted_sid, formatted_mask)


def _debug_remove_ace(sid, mask):
    """Logs info about ACE being removed.

    Arguments:
        sid: SID object.
        mask: Rights.
    """
    logger = getLogger(__name__)
    if logger.isEnabledFor(DEBUG):
        formatted_sid = _format_ace_sid(sid)
        formatted_mask = _format_ace_mask(mask)
        logger.debug("remove access for %s (%s)", formatted_sid, formatted_mask)


def _format_ace_sid(sid):
    r"""Formats the specified SID.

    Arguments:
        sid: SID object.

    Returns:
        str: Formatted SID (eg 'BUILTIN\Users').
    """
    name, domain, account_type = LookupAccountSid("", sid)
    return f"{domain}\\{name}" if domain else name


def _format_ace_mask(mask):
    """Formats the specified bit mask.

    Arguments:
        mask (int): Bit mask.

    Returns:
        str: Formatted bit mask (eg '0x1a').
    """
    if mask < 0:
        mask = 0x100000000 + mask
    return f"0x{mask:x}"



# -*- coding: utf-8 -*-
"""Pro Custodibus multifactor authentication."""

import sys
from datetime import datetime
from getpass import getpass
from os import fdopen
from re import fullmatch

from tabulate import tabulate

from procustodibus_agent.agent import ping
from procustodibus_agent.mfa.api import (
    check_mfa_api,
    do_mfa_api,
    list_mfa_api,
    login_api_with_password,
)

TZ = None


def check_mfa(cnf, endpoint, output=None):
    """Checks if MFA for specified endpoint has expired.

    Arguments:
        cnf (Config): Config object.
        endpoint (str): Endpoint ID.
        output (IOBase): Output stream.
    """
    state = check_mfa_api(cnf, endpoint).strip()
    # print to stdout or designated file
    print(state, file=output or sys.stdout)  # noqa: T201


def list_mfa(cnf, output=None):
    """Lists MFA state for all endpoints of the configured host.

    Arguments:
        cnf (Config): Config object.
        output (IOBase): Output stream.
    """
    json = list_mfa_api(cnf)
    rows = _transform_rotation_list(json)
    _print_rotation_list(rows, output)


def _transform_rotation_list(json):
    """Transforms JSON from MFA list API to shallow rows of data.

    Arguments:
        json (dict): Full API result.

    Returns:
        list: List of shallow rows with rotation data.
    """
    return [_transform_rotation_item(x, json) for x in json["data"]]


def _transform_rotation_item(item, json):
    """Transforms JSON data item from MFA list API to shallow dict of data.

    Arguments:
        item (dict): Individual data item from API.
        json (dict): Full API result.

    Returns:
        dict: Shallow dict of data for an endpoint.
    """
    attributes = item.get("attributes", {})
    relationships = item.get("relationships", {})
    interface = _find_name_of_related(relationships.get("local_interface"), json)
    peer = _find_name_of_related(relationships.get("remote_peer"), json)
    endpoint = _find_id_of_related(relationships.get("local_endpoint"))
    return {
        "due": attributes.get("due"),
        "rotated": attributes.get("rotated"),
        "sync": attributes.get("sync"),
        "interface": interface,
        "peer": peer,
        "endpoint": endpoint,
    }


def _find_id_of_related(related):
    """Finds the ID of an included item.

    Arguments:
        related (dict): Related data item from API.

    Returns:
        str: ID or `None`.
    """
    if related:
        return related["data"][0]["id"]
    return None


# simpler for now to keep this logic together rather than subdivide it more functions
def _find_name_of_related(related, json):  # noqa: CCR001
    """Finds the name of an included item from a related item.

    Arguments:
        related (dict): Related data item from API.
        json (dict): Full API result.

    Returns:
        str: Name or `None`.
    """
    if related:
        for r in related["data"]:
            for i in json.get("included", []):
                if r.get("id") == i.get("id") and r.get("type") == i.get("type"):
                    return i.get("attributes", {}).get("name")
    return None


def _print_rotation_list(rows, output=None):
    """Prints transformed, shallow data from MFA list API.

    Arguments:
        rows (list): List of transformed shallow rows.
        output (IOBase): Output stream.
    """
    values = [
        [
            row["interface"],
            row["peer"],
            row["endpoint"],
            _format_rotation_state(row) or "-",
        ]
        for row in rows
    ]

    table = tabulate(values, ["Interface", "Endpoint", "ID", "State"], "simple")
    # print to stdout or designated file
    print(table, file=output or sys.stdout)  # noqa: T201


def do_mfa(cnf, endpoint, user, password, secondary_code=None, output=None):
    """Authenticates with specified user to sync the MFA key of the specified endpoint.

    Arguments:
        cnf (Config): Config object.
        endpoint (str): Endpoint ID.
        user (str): User ID.
        password (str): User password.
        secondary_code (str): Optional secondary verification code.
        output (IOBase): Output stream.
    """
    token = login_api_with_password(cnf, user, password, secondary_code)
    do_mfa_api(cnf, endpoint, user, token)
    ping(cnf)
    json = do_mfa_api(cnf, endpoint, user, token)
    _display_mfa_results(json, output)


def read_password(password=None, file_descriptor=None, prompt=False):
    """Reads a password.

    Returns the password if `password` directly specified; or if not,
    reads from the file descriptor if `file_descriptor` specified; or if not,
    prompts for the password if `prompt` is truthy; or if not,
    returns None.

    Arguments:
        password (str): Actual password.
        file_descriptor: File descriptor number.
        prompt: True (or prompt string) to prompt for password.

    Returns:
        str: Password or None.
    """
    if password:
        return password.strip()
    if file_descriptor:
        if fullmatch(r"\d+", str(file_descriptor)):
            with fdopen(int(file_descriptor)) as f:
                return f.readline().strip()
        else:
            with open(file_descriptor) as f:
                return f.readline().strip()
    if isinstance(prompt, str):
        return getpass(prompt)
    if prompt:
        return getpass()
    return None


def _display_mfa_results(json, output=None):
    """Prints result of syncing the MFA key via API.

    Arguments:
        json (dict): API result.
        output (IOBase): Output stream.
    """
    item = json["data"][0]["attributes"]
    content = _format_rotation_state(item)

    # print to stdout or designated file
    print(content, file=output or sys.stdout)  # noqa: T201


def _format_rotation_state(item):
    """Extracts and formats the rotation state from the specified shallow data item.

    Arguments:
        item (dict): Shallow data item (ie attributes dict).

    Returns:
        str: Rotation state string or blank.
    """
    sync = item.get("sync") or item.get("preshared_key_sync")
    due = item.get("due") or item.get("preshared_key_due")
    if due:
        due = datetime.strptime(due.replace("Z", "+0000"), "%Y-%m-%dT%H:%M:%S%z")
    return _format_rotation_sync_and_due(sync, due)


# simpler for now to keep this logic together rather than subdivide it more functions
def _format_rotation_sync_and_due(sync, due):  # noqa: CCR001
    """Formats the rotation state from the specified sync data.

    Arguments:
        sync (str): Synchronization state (eg 'mismatch_pending').
        due (datetime): Synchronization due date.

    Returns:
        str: Rotation state string or blank.
    """
    if due:
        now = datetime.now().astimezone()
        if sync == "mismatch" and due:
            return "EXPIRED"
        if (not sync or sync == "match") and due > now:
            return f"OK until {due.astimezone(TZ):%x %X %Z}"
        if (not sync or sync == "match") and due < now:
            return "OK but rotation imminent"
    if sync:
        return sync.upper()
    return ""



# -*- coding: utf-8 -*-
"""Executor system commands."""

from logging import DEBUG, INFO, getLogger, root
from pathlib import Path
from platform import system
from re import match, search, sub
from subprocess import PIPE, STDOUT, CompletedProcess, run  # noqa: S404
from time import sleep

from procustodibus_agent.wg_cnf import (
    delete_wg_cnf,
    find_wg_cnf_path,
    rename_wg_cnf,
    update_wg_cnf,
)


def calculate_ip_family(address):
    """Determines whether the specified IP address is IPv4 or IPv6.

    Arguments:
        address (str): IP address (eg 'fc00:0:0:1::').

    Returns:
        int: 4 or 6.
    """
    return 6 if search(r":.*:", address) else 4


def resolvconf_interface_prefix(file="/etc/resolvconf/interface-order"):
    """Returns the prefix to append to an interface name for resolvconf commands.

    Arguments:
        file (str): Resolvconf interface-order file path to check.

    Returns:
        str: Prefix (eg 'tun') or blank ('').
    """
    path = Path(file)
    if path.exists():
        with open(path) as f:
            for line in f:
                interface_match = match(r"([A-Za-z0-9-]+)\*", line)
                if interface_match:
                    return interface_match[1]
    return ""


def get_sys_cmd():
    """Returns the command object appropriate for the current operating system.

    Returns:
        SysCmd: Command object.
    """
    s = system()
    if s == "Linux":
        return LinuxSysCmd()
    elif s == "Windows":
        return WindowsSysCmd()
    return SysCmd()


class SysCmd:
    """Runs commands for a generic operating system."""

    def start_service(self, cnf, dev, results):
        """Starts the specified WireGuard interface.

        Arguments:
            cnf (Config): Config object.
            dev (str): Interface name (eg 'wg0').
            results (list): List to append the results of the command to.
        """
        self._do_start_service(cnf, dev, results)

    def stop_service(self, cnf, dev, results):
        """Stops the specified WireGuard interface.

        Arguments:
            cnf (Config): Config object.
            dev (str): Interface name (eg 'wg0').
            results (list): List to append the results of the command to.
        """
        self._do_stop_service(cnf, dev, results)

    def update_address(self, action, address, dev, results):
        """Adds or remove the specified interface address.

        Arguments:
            action (str): "add" or "del".
            address (str): IP address or CIDR (eg '10.0.0.0/24').
            dev (str): Interface name (eg 'wg0').
            results (list): List of completed proccess objects (to add to).
        """
        self._do_update_address(action, address, dev, results)

    def update_mtu(self, mtu, dev, results):
        """Sets the specified interface MTU.

        Arguments:
            mtu (int): New MTU value (eg 1420).
            dev (str): Interface name (eg 'wg0').
            results (list): List of completed proccess objects (to add to).
        """
        self._do_update_mtu(mtu, dev, results)

    def update_dns(self, dns, search, dev, results):
        """Sets the specified interface DNS resolvers and search suffixes.

        Arguments:
            dns (list): List of DNS resolver IP addresses (eg ['9.9.9.9']).
            search (list): List of DNS search suffixes (eg ['wg.corp']).
            dev (str): Interface name (eg 'wg0').
            results (list): List of completed proccess objects (to add to).
        """
        self._do_update_dns(dns, search, dev, results)

    def check_route(self, to, dev, table, results):
        """Checks if the specified route exists.

        Arguments:
            to (str): IP address or CIDR (eg '10.0.0.0/24').
            dev (str): Interface name (eg 'wg0').
            table (str): Route table name (eg 'auto').
            results (list): List of completed proccess objects (to add to).

        Returns:
            str: None if error; 'found' if found; 'missing' if missing.
        """
        return self._do_check_route(to, dev, table, results)

    def update_route(self, action, to, dev, table, results):
        """Adds or removes the specified route.

        Arguments:
            action (str): "add" or "del".
            to (str): IP address or CIDR (eg '10.0.0.0/24').
            dev (str): Interface name (eg 'wg0').
            table (str): Route table name (eg 'auto').
            results (list): List of completed proccess objects (to add to).
        """
        check = self.check_route(to, dev, table, results)
        if (action == "add" and check == "missing") or (
            action == "del" and check == "found"
        ):
            self._do_update_route(action, to, dev, table, results)

    def set_interface(self, cnf, dev, setting, value, results):
        """Sets the specified wg setting for the specified interface.

        Arguments:
            cnf (Config): Config object.
            dev (str): Interface name (eg 'wg0').
            setting (str): Setting name (eg 'allowed-ips').
            value (str): Setting value (eg '10.0.0.1/24,fc00:0:0:1::/64').
            results (list): List to append the results of the command to.
        """
        results.append(_cmd(cnf.wg, "set", dev, setting, value))

    # wg command requires all these arguments
    def set_peer(self, cnf, dev, pk, setting, value, results):  # noqa: CFQ002
        """Sets the specified wg setting for the specified peer.

        Arguments:
            cnf (Config): Config object.
            dev (str): Interface name (eg 'wg0').
            pk (str): Peer public key.
            setting (str): Setting name (eg 'allowed-ips').
            value (str): Setting value (eg '10.0.0.1/24,fc00:0:0:1::/64').
            results (list): List to append the results of the command to.
        """
        if setting == "remove":
            results.append(_cmd(cnf.wg, "set", dev, "peer", pk, setting))
        else:
            results.append(_cmd(cnf.wg, "set", dev, "peer", pk, setting, value))

    def check_wg_cnf(self, cnf, dev, results):
        """Checks if the specified interface config file exists.

        Arguments:
            cnf (Config): Config object.
            dev (str): Name of interface to update (eg 'wg0').
            results (list): List of completed proccess objects (to add to).

        Returns:
            str: None if error; 'found' if found; 'missing' if missing.
        """
        return "found" if find_wg_cnf_path(cnf, dev).exists() else "missing"

    def create_wg_cnf(self, cnf, dev, interface, peers, results):
        """Creates the interface cnf file with the specified properties.

        Arguments:
            cnf (Config): Config object.
            dev (str): Name of interface to update (eg 'wg0').
            interface (dict): Properties of interface to update.
            peers (list): List of dicts with peer properties to update.
            results (list): List of completed proccess objects (to add to).
        """
        self.update_wg_cnf(cnf, dev, interface, peers, results)

    def update_wg_cnf(self, cnf, dev, interface, peers, results):
        """Updates the interface cnf file with the specified properties.

        Arguments:
            cnf (Config): Config object.
            dev (str): Name of interface to update (eg 'wg0').
            interface (dict): Properties of interface to update.
            peers (list): List of dicts with peer properties to update.
            results (list): List of completed proccess objects (to add to).
        """
        description = f"procustodibus_agent setconf {dev} {find_wg_cnf_path(cnf, dev)}"
        results.append(
            _as_cmd(description, lambda: update_wg_cnf(cnf, dev, interface, peers))
        )

    def rename_wg_cnf(self, cnf, old_dev, new_dev, results):
        """Renames the specified interface config file.

        Arguments:
            cnf (Config): Config object.
            old_dev (str): Old interface name (eg 'wg0').
            new_dev (str): New interface name (eg 'wg0').
            results (list): List of completed proccess objects (to add to).
        """
        description = (
            f"mv {find_wg_cnf_path(cnf, old_dev)} {find_wg_cnf_path(cnf, new_dev)}"
        )
        results.append(
            _as_cmd(description, lambda: rename_wg_cnf(cnf, old_dev, new_dev))
        )

    def delete_wg_cnf(self, cnf, dev, results):
        """Deletes the specified interface config file.

        Arguments:
            cnf (Config): Config object.
            dev (str): Interface name (eg 'wg0').
            results (list): List of completed proccess objects (to add to).
        """
        description = f"rm {find_wg_cnf_path(cnf, dev)}"
        results.append(_as_cmd(description, lambda: delete_wg_cnf(cnf, dev)))

    def _do_start_service(self, cnf, dev, results):
        raise NotImplementedError

    def _do_stop_service(self, cnf, dev, results):
        raise NotImplementedError

    def _do_update_address(self, address, dev, results):
        raise NotImplementedError

    def _do_update_mtu(self, mtu, dev, results):
        raise NotImplementedError

    def _do_update_dns(self, dns, search, dev, results):
        raise NotImplementedError

    def _do_check_route(self, to, dev, table, results):
        raise NotImplementedError

    def _do_update_route(self, action, to, dev, table, results):
        raise NotImplementedError


class LinuxSysCmd(SysCmd):
    """Runs commands for Linux."""

    def _do_start_service(self, cnf, dev, results):
        cnf_name = _find_wg_quick_cnf_name(cnf, dev)
        if cnf.manager == "systemd" and not cnf_name.startswith("/"):
            # stop first, and ignore shutdown output/errors
            _cmd("systemctl", "stop", f"wg-quick@{dev}.service")
            results.append(_cmd("systemctl", "start", f"wg-quick@{dev}.service"))
            results.append(_cmd("systemctl", "enable", f"wg-quick@{dev}.service"))
        else:
            # stop first, and ignore shutdown output/errors
            _cmd(cnf.wg_quick, "down", cnf_name)
            results.append(_cmd(cnf.wg_quick, "up", cnf_name))

    def _do_stop_service(self, cnf, dev, results):
        cnf_name = _find_wg_quick_cnf_name(cnf, dev)
        if cnf.manager == "systemd" and not cnf_name.startswith("/"):
            results.append(_cmd("systemctl", "stop", f"wg-quick@{dev}.service"))
            results.append(_cmd("systemctl", "disable", f"wg-quick@{dev}.service"))
        else:
            results.append(_cmd(cnf.wg_quick, "down", cnf_name))

    def _do_update_address(self, action, address, dev, results):
        results.append(_cmd(*_ip_route_args("address", action, address, dev)))

    def _do_update_mtu(self, mtu, dev, results):
        results.append(_cmd("ip", "link", "set", "mtu", str(mtu), "up", "dev", dev))

    def _do_update_dns(self, dns, search, dev, results):
        resolvconf_name = f"{resolvconf_interface_prefix()}{dev}"
        results.append(_cmd("resolvconf", "-d", resolvconf_name, "-f"))
        resolvconf_input = []

        if dns:
            resolvconf_input.append(f"nameserver {','.join(dns)}\n")
        if search:
            resolvconf_input.append(f"search {','.join(search)}\n")
        if resolvconf_input:
            results.append(
                _cmd(
                    "resolvconf",
                    "-a",
                    resolvconf_name,
                    "-m",
                    "0",
                    "-x",
                    stdin="".join(resolvconf_input),
                )
            )

    def _do_update_route(self, action, to, dev, table, results):
        results.append(_cmd(*_ip_route_args("route", action, to, dev, table)))

    def _do_check_route(self, to, dev, table, results):
        completed = _cmd(*_ip_route_args("route", "show", to, dev, table), quiet=True)

        if completed.returncode:
            results.append(completed)
            return None

        return "found" if completed.stdout.strip() else "missing"


class WindowsSysCmd(SysCmd):
    """Runs commands for Windows."""

    def _do_start_service(self, cnf, dev, results):
        if _sc_query_state(cnf, dev) == "RUNNING":
            results.append(_cmd("sc", "stop", _sc_service_name(cnf, dev)))
            _sc_query_state_wait_until("STOPPED", cnf, dev, results)

        cnf_path = find_wg_cnf_path(cnf, dev)
        results.append(_cmd("wireguard", "/installtunnelservice", str(cnf_path)))
        _sc_query_state_wait_until("RUNNING", cnf, dev, results)

    def _do_stop_service(self, cnf, dev, results):
        results.append(_cmd("wireguard", "/uninstalltunnelservice", dev))
        _sc_query_state_wait_until(None, cnf, dev, results)

    def _do_update_address(self, action, address, dev, results):
        args = _netsh_interface_args(address) + [
            action,
            "address",
            f"interface={dev}",
            f"address={address}",
            "store=active",
        ]
        if args[2] == "ipv4":
            args[5] = f"name={dev}"
            if action == "del":
                address = sub(r"/.*", "", address)
                args[6] = f"address={address}"
        results.append(_cmd(*args))

    def _do_update_mtu(self, mtu, dev, results):
        args = _netsh_interface_args(4) + [
            "set",
            "interface",
            f"interface={dev}",
            f"mtu={mtu}",
            "store=active",
        ]
        results.append(_cmd(*args))
        args[2] = "ipv6"
        results.append(_cmd(*args))

    def _do_update_dns(self, dns, search, dev, results):
        # TODO: search/suffix list not exposed via netsh?
        existing_4 = _netsh_show_dns_servers(4, dev, results)
        existing_6 = _netsh_show_dns_servers(6, dev, results)
        for address in existing_4 + existing_6:
            args = _netsh_interface_args(address) + [
                "del",
                "dnsservers",
                f"name={dev}",
                f"address={address}",
            ]
            results.append(_cmd(*args))
        for address in dns:
            args = _netsh_interface_args(address) + [
                "add",
                "dnsservers",
                f"name={dev}",
                f"address={address}",
            ]
            results.append(_cmd(*args))

    def _do_update_route(self, action, to, dev, table, results):
        args = _netsh_interface_args(to) + [
            action,
            "route",
            f"prefix={to}",
            f"interface={dev}",
            "store=active",
        ]
        results.append(_cmd(*args))

    def _do_check_route(self, to, dev, table, results):
        args = _netsh_interface_args(to) + ["show", "route"]
        completed = _cmd(*args, quiet=True)

        if completed.returncode:
            results.append(completed)
            return None

        for line in completed.stdout.split("\n"):
            parts = line.split()
            if len(parts) >= 6 and parts[3] == to and parts[5] == dev:
                return "found"

        return "missing"


class WireSockSysCmd(SysCmd):
    """Runs commands for WireSock."""

    def set_interface(self, cnf, dev, setting, value, results):  # noqa: D102
        pass

    # wg command requires all these arguments
    def set_peer(self, cnf, dev, pk, setting, value, results):  # noqa: CFQ002 D102
        pass

    def _do_start_service(self, cnf, dev, results):
        state = _sc_query_state(cnf, dev)
        if not state:
            results.append(
                _cmd(
                    cnf.wg,
                    "install",
                    "-start-type",
                    "2",
                    "-config",
                    str(find_wg_cnf_path(cnf, dev)),
                    "-log-level",
                    _sc_service_log_level(cnf),
                    "-lac",
                )
            )
            state = "STOPPED"

        name = _sc_service_name(cnf, dev)
        if state == "RUNNING":
            results.append(_cmd("sc", "stop", name))
            _sc_query_state_wait_until("STOPPED", cnf, dev, results)

        results.append(_cmd("sc", "start", name))
        _sc_query_state_wait_until("RUNNING", cnf, dev, results)

    def _do_stop_service(self, cnf, dev, results):
        results.append(_cmd(cnf.wg, "uninstall"))
        _sc_query_state_wait_until(None, cnf, dev, results)

    def _do_update_address(self, action, address, dev, results):
        pass

    def _do_update_mtu(self, mtu, dev, results):
        pass

    def _do_update_dns(self, dns, search, dev, results):
        pass

    def _do_update_route(self, action, to, dev, table, results):
        pass

    def _do_check_route(self, to, dev, table, results):
        return "missing"


# simpler to keep this logic together rather than subdivide it more functions
def _cmd(*args, stdin=None, quiet=False):  # noqa: CCR001
    """Executes a command with the specified arguments.

    Arguments:
        *args (list): Command arguments.
        stdin (str): Optional command input.
        quiet (bool): True to skip logging if ok (default False).

    Returns:
        Completed process object.
    """
    try:
        if stdin:
            completed = run(  # noqa: S603 S607
                args,
                input=stdin,
                stdout=PIPE,
                stderr=STDOUT,
                timeout=30,
                universal_newlines=True,
            )
        else:
            completed = run(  # noqa: S603 S607
                args, stdout=PIPE, stderr=STDOUT, timeout=30, universal_newlines=True
            )
    except Exception as e:
        completed = CompletedProcess([], 1, stdout=str(e))

    output = " ".join(args)
    stdout = completed.stdout.strip() if completed.stdout else ""
    if stdout:
        output = f"{output}\n{stdout}"

    if completed.returncode and quiet != "silent":
        getLogger(__name__).warning("err # %s", output)
    elif not quiet:
        getLogger(__name__).info("ok  # %s", output)

    return completed


def _as_cmd(description, function):
    """Runs the specified function and returns a CompletedProcess object as the result.

    Arguments:
        description (str): Command description (eg 'update config file').
        function (lambda): Zero-argument function to run.

    Returns:
        Completed process object.
    """
    try:
        function()
        getLogger(__name__).info("ok  # %s", description)
        return CompletedProcess([], 0)
    except Exception as e:
        getLogger(__name__).warning("err # %s", description, exc_info=True)
        return CompletedProcess([], 1, stdout=str(e))


def _update_wg_cnf_as_cmd(cnf, name, interface, peers=None):
    """Updates the cnf file for the specified interface with the specified properties.

    Arguments:
        cnf (Config): Config object.
        name (str): Name of interface to update (eg 'wg0').
        interface (dict): Properties of interface to update.
        peers (list): List of dicts with peer properties to update.

    Returns:
        Completed process object.
    """
    description = f"procustodibus_agent setconf {name} {find_wg_cnf_path(cnf, name)}"
    return _as_cmd(description, lambda: update_wg_cnf(cnf, name, interface, peers))


def _find_wg_quick_cnf_name(cnf, dev):
    """Finds name or path to use for wg-quick for the specified interface.

    Arguments:
        cnf (Config): Config object.
        dev (str): Name of interface (eg 'wg0').

    Returns:
        str: Name or path (eg 'wg0' or '/usr/local/etc/wireguard/wg.conf').
    """
    full_name = str(find_wg_cnf_path(cnf, dev))
    return dev if full_name == f"/etc/wireguard/{dev}.conf" else full_name


def _ip_route_args(cmd, action, to, dev, table="auto"):
    """Builds arg list for ip address/route add/del/show command.

    Arguments:
        cmd (str): "address" or "route".
        action (str): "add", "del", or "show".
        to (str): IP address or CIDR (eg '10.0.0.0/24').
        dev (str): Interface name (eg 'wg0').
        table (str): Route table name (eg 'auto').

    Returns:
        list: Arg list.
    """
    family = calculate_ip_family(to)

    args = ["ip", f"-{family}", cmd, action, to, "dev", dev]
    if table != "auto":
        args += ["table", table]

    return args


def _netsh_interface_args(to):
    """Builds arg list for netsh route add/del/show command.

    Arguments:
        to (str): IP address or CIDR (eg '10.0.0.0/24').

    Returns:
        list: Arg list.
    """
    family = to if to in (4, 6) else calculate_ip_family(to)
    return ["netsh", "interface", f"ipv{family}"]


def _netsh_show_dns_servers(family, dev, results):
    """Runs netsh command to show DNS servers.

    Arguments:
        family (int): 4 or 6.
        dev (str): Interface name (eg 'wg0').
        results (list): List of completed proccess objects (to add to).

    Returns:
        list: List of DNS servers.
    """
    args = _netsh_interface_args(family) + ["show", "dnsservers", dev]
    completed = _cmd(*args, quiet=True)
    servers = []

    for line in completed.stdout.split("\n"):
        m = search(r"(\d+\.\d+\.\d+\.\d+|[\da-f]+:[\da-f.:%]+)", line)
        if m:
            servers.append(m.group(0))

    return servers


def _sc_query_state_wait_until(state, cnf, dev, results=None):
    """Runs sc command to query service status until it matches specified state.

    Arguments:
        state (str): Desired state (eg 'STOPPED').
        cnf (Config): Config object.
        dev (str): Interface name (eg 'wg0').
        results (list): List of completed proccess objects (to add to).

    Returns:
        bool: True if reached specified state.
    """
    for x in range(0, 5):
        sleep(x / 4)
        if _sc_query_state(cnf, dev) == state:
            return True

    sleep(2)
    if _sc_query_state(cnf, dev, results) == state:
        return True
    return False


def _sc_query_state(cnf, dev, results=None):
    """Runs sc command to query service status.

    Arguments:
        cnf (Config): Config object.
        dev (str): Interface name (eg 'wg0').
        results (list): List of completed proccess objects (to add to).

    Returns:
        str: Service state (eg 'RUNNING'), or None if service unknown.
    """
    completed = _cmd("sc", "query", _sc_service_name(cnf, dev), quiet="silent")
    if results is not None:
        results.append(completed)

    for line in completed.stdout.split("\n"):
        parts = line.split()
        if len(parts) >= 4 and parts[0] == "STATE":
            return parts[3]
    return None


def _sc_service_name(cnf, dev):
    """Returns service name.

    Arguments:
        cnf (Config): Config object.
        dev (str): Interface name (eg 'wg0').

    Returns:
        str: Service name (eg 'wiresock-client-service').
    """
    if cnf.wiresock:
        return "wiresock-client-service"
    return f"WireGuardTunnel${dev}"


def _sc_service_log_level(cnf):
    """Returns service log level.

    Arguments:
        cnf (Config): Config object.

    Returns:
        str: Service log level (eg 'debug').
    """
    level = root.level
    if level < DEBUG:
        return "all"
    if level == DEBUG:
        return "debug"
    if level <= INFO:
        return "info"
    return "none"



# -*- coding: utf-8 -*-
"""cx_Freeze Win32Service for the agent."""

import sys
from pathlib import Path

# cx_Logging only needs to be installed in the build_exe environment
try:
    import cx_Logging  # noqa: I900
except ImportError:
    cx_Logging = None  # noqa: N816

from procustodibus_agent.agent import ping_loop
from procustodibus_agent.cnf import Cnf
from procustodibus_agent.executor.execution import set_sys_cmd
from procustodibus_agent.executor.sys_cmd import WireSockSysCmd
from procustodibus_agent.windows.cnf import restrict_access_to_cnf_dir

LOG_LEVEL = "INFO"
LOOP_SECONDS = 120


class Service:
    """cx_Freeze Win32Service for the agent."""

    def initialize(self, cnf_file):
        """Called when the service is starting.

        Arguments:
            cnf_file (str): Path to configuration file.
        """
        try:
            init_log()
            cnf_file = normalize_cnf_file(cnf_file)
            self.cnf = Cnf(cnf_file, LOG_LEVEL, LOOP_SECONDS)
            init_wiresock(self.cnf)
        except Exception:
            cx_Logging.LogException()

    def run(self):
        """Called when the service is running."""
        init_args = [self.cnf.cnf_file, LOG_LEVEL, LOOP_SECONDS]
        try:
            restrict_access_to_cnf_dir(self.cnf)
            ping_loop(self.cnf, init_args)
        except Exception:
            cx_Logging.LogException()

    def stop(self):
        """Called when the service is stopping."""
        self.cnf.loop = 0


def init_log():
    """Initializes service logging."""
    log_dir = get_log_dir()
    cx_Logging.StartLogging(str(log_dir / "init.log"), cx_Logging.DEBUG)
    sys.stdout = open(log_dir / "stdout.log", "a")


def get_log_dir():
    """Gets service logging directory.

    Returns:
        str: Path to logging directory.
    """
    executable_dir = Path(sys.executable).parent
    log_dir = executable_dir / "log"
    Path.mkdir(log_dir, parents=True, exist_ok=True)
    return log_dir


def normalize_cnf_file(cnf_file):
    """Checks that specified configuration file has the correct file extension.

    Arguments:
        cnf_file (str): Path to configuration file.

    Returns:
        str: Path to configuration file or blank.
    """
    file = cnf_file if cnf_file and cnf_file.endswith(".conf") else ""
    cx_Logging.Debug("service config is %r", file)
    return file


def init_wiresock(cnf):
    """Initializes WireSock if necessary.

    Arguments:
        cnf (Config): Config object.
    """
    if cnf.wiresock:
        set_sys_cmd(WireSockSysCmd())



# -*- coding: utf-8 -*-
"""Routing and addressing utilities."""

import time
from ipaddress import IPv4Address, IPv4Interface
from itertools import zip_longest
from json import loads
from platform import system
from re import fullmatch, match, sub
from subprocess import PIPE, run  # noqa: S404

FAKE_WIRESOCK_HANDSHAKES = {}


def run_ip_address_show(device):
    """Runs `ip address show` command for specified device.

    Arguments:
        device (str): Device name (eg 'wg0').

    Returns:
        dict: Parsed JSON from `ip --json address show` command.
    """
    args = ["ip", "--json", "address", "show", "dev", device]
    result = run(args, stdout=PIPE)  # noqa: S603 S607
    output = result.stdout.decode("utf-8")
    return loads(output)[0] if fullmatch(r"\[\{.*\}\]\s*", output) else {}


def run_netsh_show_addresses(device):
    """Runs `netsh` addresses command for specified device.

    Arguments:
        device (str): Device name (eg 'wg0').

    Returns:
        list: Lines from `netsh` command.
    """
    ipv4 = _run_netsh_show_addresses_for_family(device, "ipv4")
    ipv6 = _run_netsh_show_addresses_for_family(device, "ipv6")
    return ipv4 + ipv6


def _run_netsh_show_addresses_for_family(device, family):
    """Runs `netsh` addresses command for specified device and IP version.

    Arguments:
        device (str): Device name (eg 'wg0').
        family (str): IP family ('ipv4' or 'ipv6')

    Returns:
        list: Lines from `netsh` command.
    """
    args = ["netsh", "interface", family, "show", "addresses", device]
    result = run(args, stdout=PIPE)  # noqa: S603 S607
    output = result.stdout.decode("utf-8")
    return [x.strip() for x in output.split("\n")]


def run_ifconfig(device):
    """Runs `ifconfig` command for specified device.

    Arguments:
        device (str): Device name (eg 'wg0').

    Returns:
        list: Lines from `ifconfig` command.
    """
    args = ["ifconfig", device]
    result = run(args, stdout=PIPE)  # noqa: S603 S607
    output = result.stdout.decode("utf-8")
    return [x.strip() for x in output.split("\n")]


def annotate_wg_show_with_ip_address_show(interfaces):
    """Annotates parsed output of `wg show` with output of `ip address show`.

    Arguments:
        interfaces (dict): Dict parsed from `wg show` command.

    Returns:
        dict: Same dict with additional properties.
    """
    for name, properties in interfaces.items():
        _annotate_interface(name, properties)
        properties["up"] = True if properties.get("address") else False
    return interfaces


def _annotate_interface(name, properties):
    """Annotates specified dict with wg config for specified interface.

    Arguments:
        name (str): Interface name (eg 'wg0').
        properties (dict): Dict of interface properties.

    Returns:
        dict: Same dict with additional properties.
    """
    if system() == "Linux":
        _annotate_interface_with_ip_address_show(name, properties)
    elif system() == "Windows":
        _annotate_interface_with_netsh_show_addresses(name, properties)
    else:
        _annotate_interface_with_ifconfig(name, properties)
    return properties


def _annotate_interface_with_ip_address_show(name, properties):
    """Annotates specified dict with wg config for specified interface.

    Arguments:
        name (str): Interface name (eg 'wg0').
        properties (dict): Dict of interface properties.

    Returns:
        dict: Same dict with additional properties.
    """
    info = run_ip_address_show(name)
    if info:
        properties["address"] = [_format_address_info(a) for a in info["addr_info"]]
    return properties


def _format_address_info(info):
    """Formats the addr_info object from `ip address show` as a CIDR.

    Arguments:
        info (dict): addr_info object.

    Returns:
        string: CIDR.
    """
    return "{}/{}".format(info["local"], info["prefixlen"])


def _annotate_interface_with_netsh_show_addresses(name, properties):
    """Annotates specified dict with wg config for specified interface.

    Arguments:
        name (str): Interface name (eg 'wg0').
        properties (dict): Dict of interface properties.

    Returns:
        dict: Same dict with additional properties.
    """
    output = run_netsh_show_addresses(name)
    addresses = list(filter(None, [_parse_netsh_address(x) for x in output]))

    if addresses:
        prefixes = list(filter(None, [_parse_netsh_prefix_length(x) for x in output]))
        combined = [f"{a}/{p}" if p else a for a, p in zip_longest(addresses, prefixes)]
        properties["address"] = combined

    return properties


def _parse_netsh_address(line):
    """Parses the ip address out of the specified netsh output line.

    Arguments:
        line (str): netsh output line (eg 'IP Address:    10.0.0.1').

    Returns:
        str: Address or empty string (eg '10.0.0.1').
    """
    m = match(r"(?:IP )?Address:?\s+([\da-f.:%]+)", line)
    return m.group(1) if m else ""


def _parse_netsh_prefix_length(line):
    """Parses the subnet prefix length out of the specified netsh output line.

    Arguments:
        line (str): netsh output line (eg 'Subnet Prefix:    10.0.0.0/24').

    Returns:
        int: Prefix or None (eg 0).
    """
    m = match(r"Subnet Prefix:\s+[\da-f.:%]+/(\d+).*", line)
    return int(m.group(1)) if m else None


def _annotate_interface_with_ifconfig(name, properties):
    """Annotates specified dict with wg config for specified interface.

    Arguments:
        name (str): Interface name (eg 'wg0').
        properties (dict): Dict of interface properties.

    Returns:
        dict: Same dict with additional properties.
    """
    addresses = list(filter(None, [_parse_inet_address(x) for x in run_ifconfig(name)]))
    if addresses:
        properties["address"] = addresses
    return properties


def _parse_inet_address(line):
    """Parses the inet or inet6 address out of the specified ifconfig output line.

    Arguments:
        line (str): ifconfig output line (eg 'inet6 ffc0:: prefixlen 64').

    Returns:
        str: CIDR or empty string (eg 'ffc0::/64').
    """
    if not line.startswith("inet"):
        return ""

    parts = line.split()
    if len(parts) < 4:
        return ""
    elif parts[2] == "-->" and len(parts) >= 6:
        del parts[2:4]

    return _format_inet_address_parts(parts)


def _format_inet_address_parts(parts):
    """Formats the specified ifconfig output tokens as a CIDR.

    Arguments:
        parts (list): ifconfig output tokens
            (eg ['inet6', 'ffc0::', 'prefixlen', '64']).

    Returns:
        str: CIDR or empty string (eg 'ffc0::/64').
    """
    if parts[1] == "addr:":
        return _strip_interface_id(parts[2])
    elif parts[1].startswith("addr:"):
        ip = parts[1][5:]
        mask = parts[3][5:] if parts[3].startswith("Mask:") else 32
        return str(IPv4Interface((ip, mask)))
    elif parts[2] == "prefixlen":
        ip = _strip_interface_id(parts[1])
        mask = parts[3]
        return f"{ip}/{mask}"
    elif parts[2] == "netmask":
        ip = _strip_interface_id(parts[1])
        mask = str(IPv4Address(int(parts[3], 16)))
        return str(IPv4Interface((ip, mask)))
    return ""


def _strip_interface_id(ip):
    """Strips the interface number or name from the specified IP address.

    Arguments:
        ip (str): IP address (eg 'fe80::169d:99ff:fe7f:8c67%utun0').

    Returns:
        str: IP address (eg 'fe80::169d:99ff:fe7f:8c67').
    """
    return sub(r"%[^/]*", "", ip)


def annotate_wg_show_with_up(interfaces):
    """Annotates parsed output of `wg show` with 'up' property.

    Arguments:
        interfaces (dict): Dict parsed from `wg show` command.

    Returns:
        dict: Same dict with additional properties.
    """
    output = []
    for name, properties in interfaces.items():
        if not output:
            output = _run_netsh_show_interface()
        properties["up"] = _parse_up_for_interface(output, name)
    return interfaces


def _run_netsh_show_interface(device=None):
    """Runs `netsh` interface command for specified device.

    Arguments:
        device (str): Device name (eg 'wg0').

    Returns:
        list: Lines from `netsh` command.
    """
    args = ["netsh", "interface", "show", "interface"]
    if device:
        args.append(device)
    result = run(args, stdout=PIPE)  # noqa: S603 S607
    output = result.stdout.decode("utf-8")
    return [x.split() for x in output.split("\n")]


def _parse_up_for_interface(rows, device):
    """Calculates whether the specified device is up or not.

    Arguments:
        rows (list): Netsh subinterfaces output as a list of a list of columns.
        device (str): Device name (eg 'wg0').

    Returns:
        int: Total or 0.
    """
    for row in rows:
        if len(row) > 3 and row[3] == device:
            return row[0] == "Enabled" and row[1] == "Connected"
    return False


def annotate_wg_show_with_tx(interfaces):
    """Annotates parsed output of `wg show` with peer tx and rx.

    Arguments:
        interfaces (dict): Dict parsed from `wg show` command.

    Returns:
        dict: Same dict with additional properties.
    """
    output = []
    for name, properties in interfaces.items():
        peers = properties.get("peers", [])
        if len(peers) == 1 and properties.get("up"):
            if not output:
                output = run_netsh_show_subinterfaces()
            peer = list(peers.values())[0]
            peer["transfer_rx"] = _parse_total_for_subinterface(output, name, 2)
            peer["transfer_tx"] = _parse_total_for_subinterface(output, name, 3)
            _fake_wiresock_handshake(name, peer)
    return interfaces


def run_netsh_show_subinterfaces(device=None):
    """Runs `netsh` subinterfaces command for specified device.

    Arguments:
        device (str): Device name (eg 'wg0'); default all devices.

    Returns:
        list: Lines from `netsh` command.
    """
    ipv4 = _run_netsh_show_subinterfaces_for_family(device, "ipv4")
    ipv6 = _run_netsh_show_subinterfaces_for_family(device, "ipv6")
    return ipv4 + ipv6


def _run_netsh_show_subinterfaces_for_family(device, family):
    """Runs `netsh` subinterfaces command for specified device and IP version.

    Arguments:
        device (str): Device name (eg 'wg0').
        family (str): IP family ('ipv4' or 'ipv6')

    Returns:
        list: Lines from `netsh` command.
    """
    args = ["netsh", "interface", family, "show", "subinterfaces"]
    if device:
        args.append(device)
    result = run(args, stdout=PIPE)  # noqa: S603 S607
    output = result.stdout.decode("utf-8")
    return [x.split() for x in output.split("\n")]


def _parse_total_for_subinterface(rows, device, column):
    """Calculates the total of a subinterfaces column for the specified device.

    Arguments:
        rows (list): Netsh subinterfaces output as a list of a list of columns.
        device (str): Device name (eg 'wg0').
        column (int): Zero-based column (eg 2).

    Returns:
        int: Total or 0.
    """
    total = 0
    for row in rows:
        if len(row) > 4 and row[4] == device:
            total += int(row[column])
    return total


def _fake_wiresock_handshake(device, peer):
    """Fakes the latest_handshake time for WireSock.

    (So that the API will record an entry for the transfer stats.)

    Arguments:
        device (str): Device name (eg 'wg0').
        peer (dict): Peer properties to read and update.
    """
    rx = peer.get("transfer_rx")
    tx = peer.get("transfer_tx")
    # skip if device has not transferred anything ever
    if not rx and not tx:
        return

    global FAKE_WIRESOCK_HANDSHAKES
    fake = FAKE_WIRESOCK_HANDSHAKES.get(device)
    if not fake:
        fake = {"rx": 0, "tx": 0, "handshake": 0}
        FAKE_WIRESOCK_HANDSHAKES[device] = fake

    # skip if device has not transferred anything since last time
    if rx <= fake["rx"] and tx <= fake["tx"]:
        return

    handshake = int(time.time())
    # don't increment handshake if less than 2 minutes since last time
    if handshake <= fake["handshake"] + 120:
        handshake = fake["handshake"]

    fake["rx"] = rx
    fake["tx"] = tx
    fake["handshake"] = handshake
    peer["latest_handshake"] = handshake



# -*- coding: utf-8 -*-
"""Agent logic."""
from io import StringIO
from logging import getLogger
from time import sleep

from procustodibus_agent import __version__ as version
from procustodibus_agent.api import ping_api
from procustodibus_agent.cnf import reload_cnf_if_modified
from procustodibus_agent.connectivity import check_connectivity
from procustodibus_agent.executor import execute_desired
from procustodibus_agent.ip_route import (
    annotate_wg_show_with_ip_address_show,
    annotate_wg_show_with_tx,
    annotate_wg_show_with_up,
)
from procustodibus_agent.resolve_hostname import apply_endpoint_hostnames
from procustodibus_agent.wg import filter_wg_show, parse_wg_show, run_wg_show
from procustodibus_agent.wg_cnf import (
    annotate_wg_show_with_wg_cnf,
    load_all_from_wg_cnf,
)


def is_connection_good(cnf, message="Checking connectivity", expect_error=False):
    """Runs connectivity check and logs result.

    Arguments:
        cnf (Config): Config object.
        message (str): Message to log (defaults to 'Checking connectivity').
        expect_error (bool): True if error expected (defaults to False).

    Returns:
        bool: True if connection is good.
    """
    buffer = StringIO()
    # print to buffer
    print(message, file=buffer)  # noqa: T201

    error = check_connectivity(cnf, buffer)
    if error:
        if not expect_error:
            getLogger(__name__).error(buffer.getvalue())
        return False

    getLogger(__name__).info(buffer.getvalue())
    return True


# simpler to keep logic in same function even if it makes cognitive-complexity high
def ping_loop(cnf, cli_args=None):  # noqa: CCR001
    """Pings continuously as specified by looping configuration.

    Arguments:
        cnf (Config): Config object.
        cli_args (list): List of arguments to pass to Cnf constructor when reloading.
    """
    good = is_connection_good(cnf, message=f"Starting agent {version}")
    hide_error = True

    while cnf.loop:
        if good:
            try:
                ping(cnf)
            except Exception:
                getLogger(__name__).exception("ping failed")
                good = False
                hide_error = False
        else:
            good = is_connection_good(cnf, expect_error=hide_error)
            hide_error = True
            if good:
                continue
        sleep(cnf.loop)
        cnf = reload_cnf_if_modified(cnf, cli_args)


def ping(cnf):
    """Gathers wg info and pings api, executing desired changes from response.

    Arguments:
        cnf (Config): Config object.
    """
    interfaces = interrogate(cnf)
    response = ping_api(cnf, interfaces)
    executed = execute_desired(cnf, interfaces, response.get("data"))

    if executed:
        # allow time for changes to propagate
        sleep(2)
        # notify api of changes
        interfaces = interrogate(cnf)
        ping_api(cnf, interfaces, executed)
    else:
        # check for endpoint dns updates
        apply_endpoint_hostnames(cnf, interfaces)


def interrogate(cnf):
    """Extracts interface dict from wg show and wg config.

    Arguments:
        cnf (Config): Config object.

    Returns:
        dict: Interface info parsed from wg etc.
    """
    if cnf.wiresock:
        interfaces = filter_wg_show(load_all_from_wg_cnf(cnf), cnf)
        interfaces = annotate_wg_show_with_up(interfaces)
        interfaces = annotate_wg_show_with_tx(interfaces)
    else:
        interfaces = filter_wg_show(parse_wg_show(run_wg_show(cnf)), cnf)
        interfaces = annotate_wg_show_with_wg_cnf(cnf, interfaces)
        interfaces = annotate_wg_show_with_ip_address_show(interfaces)
    return interfaces



# -*- coding: utf-8 -*-
"""Utilities for updating endpoints with resolved hostnames."""

from contextlib import suppress
from logging import getLogger
from re import fullmatch, search
from socket import AF_INET, AF_INET6, IPPROTO_UDP, SOCK_DGRAM, gaierror, getaddrinfo

from procustodibus_agent.wg import run_wg_set


def resolve_endpoint_hostname(endpoint):
    """Resolves the specified endpoint to an IP address.

    Arguments:
        endpoint (str): Endpoint with a hostname (eg 'vpn.example.com:51820').

    Returns:
        tuple: Resolved endpoint, hostname, port
        (eg ('[fd01::1]:51820', 'vpn.example.com', 51820)).
    """
    hostname, port = split_endpoint_hostname_and_port(endpoint)
    if hostname:
        addresses = lookup_addresses(hostname, 0, port)
        if addresses:
            endpoint = format_endpoint_address(addresses[0], port)
        else:
            getLogger(__name__).warning(f"hostname resolution failed: {hostname}")
    return endpoint, hostname, port


def apply_endpoint_hostnames(cnf, interfaces):
    """Updates the endpoints of the specified peers if configured with a hostname.

    Arguments:
        cnf (Cnf): Config object.
        interfaces (dict): Interfaces data.
    """
    resolve = cnf.resolve_hostnames
    if resolve == "once" or cnf.wiresock:
        return

    for name, interface in interfaces.items():
        for pubkey, peer in interface.get("peers", {}).items():
            hostname = peer.get("hostname")
            endpoint = peer.get("endpoint")
            apply_endpoint_hostname(cnf, name, pubkey, hostname, resolve, endpoint)


def apply_endpoint_hostname(cnf, name, pubkey, hostname, family, endpoint):
    """Updates the specified peer's endpoint using the specified hostname.

    Arguments:
        cnf (Cnf): Config object.
        name (str): Interface name (eg 'wg0').
        pubkey (str): Peer public key.
        hostname (str): Hostname to use (eg 'foo.example.com').
        family (str): Preferred address family ('ipv4' or 'ipv6' or 'auto').
        endpoint (str): Existing endpoint (eg '10.10.10.10:51820').
    """
    if not name or not pubkey or not hostname or not endpoint:
        return

    family = get_address_family(family)
    address, port = split_endpoint_address_and_port(endpoint)
    addresses = lookup_addresses(hostname, family, port)

    if not addresses:
        getLogger(__name__).warning(f"hostname resolution failed: {hostname}")
    elif address not in addresses:
        set_endpoint_address(cnf, name, pubkey, addresses[0], port)


def set_endpoint_address(cnf, name, pubkey, address, port):
    """Runs the `wg set` command to set the endpoint of the specified peer.

    Arguments:
        cnf (Cnf): Config object.
        name (str): Interface name (eg 'wg0').
        pubkey (str): Peer public key.
        address (str): Endpoint IP address (eg '10.10.10.10').
        port (int): Endpoint port (eg 51820).
    """
    endpoint = format_endpoint_address(address, port)
    run_wg_set(cnf, [name, "peer", pubkey, "endpoint", endpoint])


def format_endpoint_address(address, port):
    """Formats the specified endpoint IP address and port.

    Arguments:
        address (str): Endpoint IP address (eg 'fd01::1').
        port (int): Endpoint port (eg 51820).

    Returns:
        str: Formatted endpoint address (eg '[fd01::1]:51820').
    """
    return f"[{address}]:{port}" if address.find(":") >= 0 else f"{address}:{port}"


def get_address_family(family):
    """Returns address family number for IP address type.

    Arguments:
        family: Address family number (AF_INET or AF_INET6)
            or string ('ipv4' or 'ipv6').

    Returns:
        Address family number.
    """
    if family == "ipv4":
        return AF_INET
    if family == "ipv6":
        return AF_INET6
    elif isinstance(family, str):
        return 0
    return family or 0


def split_endpoint_address_and_port(endpoint):
    """Splits the specified endpoint into IP address and port.

    Arguments:
        endpoint (str): IP address and port (eg '[fc00:0:0:1::]:51820').

    Returns:
        Tuple of string IP address and integer port number.
    """
    if not endpoint:
        return "", 51820

    ipv6 = fullmatch(r"\[([^\]]+)\]:(\d+)", endpoint)
    if ipv6:
        return ipv6[1], int(ipv6[2])

    ipv4 = endpoint.split(":")
    if len(ipv4) == 2:
        with suppress(ValueError):
            return ipv4[0], int(ipv4[1])

    return ipv4[0], 51820


def split_endpoint_hostname_and_port(endpoint):
    """Splits the specified endpoint into hostname and port.

    If the endpoint has an IP address instead of a hostname, returns ('', 0).

    Arguments:
        endpoint (str): Hostname and port (eg 'vpn.example.com:51820').

    Returns:
        Tuple of string hostname and integer port number.
    """
    if not endpoint or endpoint.startswith("[") or not search("[A-Za-z]", endpoint):
        return "", 0

    parts = endpoint.split(":")
    if len(parts) == 2:
        with suppress(ValueError):
            return parts[0], int(parts[1])

    return parts[0], 51820


def lookup_addresses(hostname, family=0, port=None):
    """Looks up the DNS entries for the specified hostname.

    Arguments:
        hostname (str): Hostname to lookup (eg 'foo.example.com').
        family (int): Address family (AF_NET or AF_NET6), defaults to 0 (any).
        port (int): Port number, defaults to None (any).

    Returns:
        list: List of string IP addresses.
    """
    try:
        info = getaddrinfo(hostname, port, family, SOCK_DGRAM, IPPROTO_UDP)
        return [x[4][0] for x in info]
    except gaierror:
        return []



# -*- coding: utf-8 -*-
"""cx_Freeze config for the agent as a Win32Service."""

NAME = "ProCustodibusAgent%s"
DISPLAY_NAME = "Pro Custodibus Agent %s"
MODULE_NAME = "procustodibus_agent.windows.service"
CLASS_NAME = "Service"
DESCRIPTION = "Synchronizes your WireGuard settings with Pro Custodibus."
AUTO_START = False
SESSION_CHANGES = False
